# -*- coding: utf-8 -*-

"""Provides shared mixins for ORM classes."""

__all__ = [
    "PolymorphicBaseMixin",
    "PolymorphicMixin",
    "TouchMixin",
]

import logging

from datetime import datetime

from sqlalchemy import Column
from sqlalchemy import Unicode
from sqlalchemy.ext.declarative import declared_attr

from pyramid_basemodel import save as save_to_db

logger = logging.getLogger(__name__)


class PolymorphicBaseMixin:
    """
    PolymorphicMixin streamline inheritance.

    Provides a dynamically generated ``__mapper_args__`` property for
    [single table inherited][] ORM classes::

    [single table inherited]: http://bit.ly/TBDmMx
    """

    discriminator = Column("type", Unicode(16))

    @declared_attr
    def __mapper_args__(self):
        """Set the ``polymorphic_identity`` value to the lower case class name."""
        return {"polymorphic_on": self.discriminator, "polymorphic_identity": self.__class__.__name__.lower()}


class PolymorphicMixin:
    """
    PolymorphicMixin streamline inheritance.

    Provides a dynamically generated ``__mapper_args__`` property for
    [single table inherited][] ORM classes::

    [single table inherited]: http://bit.ly/TBDmMx
    """

    @declared_attr
    def __mapper_args__(self):
        """Set the ``polymorphic_identity`` value to the lower case class name."""
        return {"polymorphic_identity": self.__class__.__name__.lower()}


class TouchMixin:
    """Provides ``touch`` and ``propagate_touch`` methods."""

    def propagate_touch(self):
        """
        Override to propagate touch events to relations.

        Note that this event *should not* be  called in response to an
        SQLAlchemy ORM attribute modified event, as you can't reliably
        update relations in an attribute event handler.
        """

    def touch(self, propagate=True, now=datetime.utcnow, save=save_to_db):
        """Update self.modified."""
        # Update self's modified date.
        self.modified = now()
        save(self)

        # Call propagate touch.
        if propagate:
            self.propagate_touch()



# -*- coding: utf-8 -*-
"""
Main pyramid_basemodel module.

Provides global scoped ``Session`` and declarative ``Base``, ``BaseMixin``
class and ``bind_engine`` function.

To use, import and, e.g.: inherit from the base classes::

  >>> class MyModel(Base, BaseMixin):
  ...     __tablename__ = 'my_model'
  ...
  >>> instance = MyModel()
  >>> Session.add(instance)
  >>> # etc.

To automatically bind the base metadata and session to your db engine, just
include the package::

  config.include('pyramid_basemodel')

"""

__version__ = "0.6.0"

__all__ = [
    "Base",
    "BaseMixin",
    "Session",
    "bind_engine",
]

import inflect
from datetime import datetime

from zope.interface import classImplements
from zope.sqlalchemy import register

from sqlalchemy import engine_from_config
from sqlalchemy import Column, DateTime, Integer
from sqlalchemy.orm import declarative_base, scoped_session, sessionmaker

from pyramid.path import DottedNameResolver
from pyramid.settings import asbool

from pyramid_basemodel.interfaces import IDeclarativeBase

Session = scoped_session(sessionmaker())
register(Session)
Base = declarative_base()
classImplements(Base, IDeclarativeBase)


class classproperty:
    """A basic [class property](http://stackoverflow.com/a/3203659)."""

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class BaseMixin:
    """
    Default Base Model Mixin.

    Provides an int ``id`` as primary key, ``version``, ``created`` and
    ``modified`` columns and a scoped ``self.query`` property.
    """

    #: primary key
    id = Column(Integer, primary_key=True)

    #: schema version
    version = Column("v", Integer, default=1)

    #: timestamp of object creation
    created = Column("c", DateTime, default=datetime.utcnow)

    #: timestamp of object's latest update
    modified = Column("m", DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

    query = Session.query_property()

    @classproperty
    def class_name(cls):
        """
        Determine class name based on the _class_name or the __tablename__.

        If provided, defaults to ``cls._class_name``, otherwise default to
        ``cls.plural_class_name``, which is derived from the cls.__tablename__.

        If singularising the plural class name doesn't work, uses the
          ``cls.__name__``
        """
        # Try the manual override.
        if hasattr(cls, "_class_name"):
            return cls._class_name

        singularise = inflect.engine().singular_noun
        name = singularise(cls.plural_class_name)
        if name:
            return name

        # If that didn't work, fallback on the class name.
        return cls.__name__

    @classproperty
    def class_slug(cls):
        """Class slug based on either _class_slug or __tablename__."""
        return getattr(cls, "_class_slug", cls.__tablename__)

    @classproperty
    def singular_class_slug(cls):
        """Return singular version of ``cls.class_slug``."""
        # If provided, use ``self._singular_class_slug``.
        if hasattr(cls, "_singular_class_slug"):
            return cls._singular_class_slug

        # Otherwise singularise the class_slug.
        if inflect is not None:
            singularise = inflect.engine().singular_noun
            slug = singularise(cls.class_slug)
            if slug:
                return slug

        # If that didn't work, fallback on the class name.
        return cls.class_name.split()[-1].lower()

    @classproperty
    def plural_class_name(cls):
        """Return plurar version of a class name."""
        # If provided, use ``self._plural_class_name``.
        if hasattr(cls, "_plural_class_name"):
            return cls._plural_class_name

        # Otherwise pluralise the literal class name.
        return cls.__tablename__.replace("_", " ").title()


def save(instance_or_instances, session=Session):
    """
    Save model instance(s) to the db.

    Both single and multiple instances can be saved.
    """
    v = instance_or_instances
    if isinstance(v, list) or isinstance(v, tuple):
        session.add_all(v)
    else:
        session.add(v)


def bind_engine(engine, session=Session, base=Base, should_create=False, should_drop=False):
    """
    Bind the ``session`` and ``base`` to the ``engine``.

    :param should_create: Triggers create tables on all models
    :param should_drop: Triggers drop on all tables
    """
    session.configure(bind=engine)
    base.metadata.bind = engine
    if should_drop:
        base.metadata.drop_all(engine)
    if should_create:
        base.metadata.create_all(engine)


def includeme(config):
    """Bind to the db engine specifed in ``config.registry.settings``."""
    # Bind the engine.
    settings = config.get_settings()
    engine_kwargs_factory = settings.pop("sqlalchemy.engine_kwargs_factory", None)
    if engine_kwargs_factory:
        kwargs_factory = config.maybe_dotted(engine_kwargs_factory)
        engine_kwargs = kwargs_factory(config.registry)
    else:
        engine_kwargs = {}
    pool_class = settings.pop("sqlalchemy.pool_class", None)
    if pool_class:
        dotted_name = DottedNameResolver()
        engine_kwargs["poolclass"] = dotted_name.resolve(pool_class)
    should_bind = asbool(settings.get("basemodel.should_bind_engine", True))
    should_create = asbool(settings.get("basemodel.should_create_all", False))
    should_drop = asbool(settings.get("basemodel.should_drop_all", False))
    if should_bind:
        engine = engine_from_config(settings, "sqlalchemy.", **engine_kwargs)
        config.action(
            None,
            bind_engine,
            (engine,),
            {"should_create": should_create, "should_drop": should_drop},
        )



# -*- coding: utf-8 -*-

"""Provides a base ORM mixin for models that need a name and a url slug."""

__all__ = [
    "BaseSlugNameMixin",
]

import logging

import slugify

from sqlalchemy import exc as sa_exc
from sqlalchemy import inspect as sa_inspect
from sqlalchemy.ext import declarative
from sqlalchemy.schema import Column
from sqlalchemy.types import Unicode

from pyramid_basemodel.util import ensure_unique
from pyramid_basemodel.util import generate_random_digest

from pyramid_basemodel import Session

logger = logging.getLogger(__name__)


class BaseSlugNameMixin:
    """
    Base mixin delivering a slug functionality.

    ORM mixin class that provides ``slug`` and ``name`` properties, with a
    ``set_slug`` method to set the slug value from the name and a default
    name aware factory classmethod.
    """

    _max_slug_length = 64
    _slug_is_unique = True

    @property
    def __name__(self):
        """Url friendly name."""
        return self.slug

    @declarative.declared_attr
    def slug(cls):
        """Get url friendly slug, e.g.: `foo-bar`."""
        return Column(Unicode(cls._max_slug_length), nullable=False, unique=cls._slug_is_unique)

    @declarative.declared_attr
    def name(cls):
        """Get human readable name, e.g.: `Foo Bar`."""
        return Column(Unicode(cls._max_slug_length), nullable=False)

    def set_slug(
        self,
        candidate=None,
        gen_digest=generate_random_digest,
        inspect=sa_inspect,
        session=Session,
        to_slug=slugify.slugify,
        unique=ensure_unique,
    ):
        """
        Generate and set a unique ``self.slug`` from ``self.name``.

        :param get_digest: function to generate random digest. Used of there's no name set.
        :param inspect:
        :param session: SQLAlchemy's session
        :param to_slug: slugify function
        :param unique: unique function
        """
        # Generate a candidate slug.
        if candidate is None:
            if self.name:
                candidate = to_slug(self.name)
            if not candidate:
                candidate = gen_digest(num_bytes=16)

        # Make sure it's not longer than 64 chars.
        length = self._max_slug_length
        length_minus_ellipsis = length - 3
        candidate = candidate[:length]
        unique_candidate = candidate[:length_minus_ellipsis]

        # If there's no name, only set the slug if its not already set.
        if self.slug and not self.name:
            return

        # If there is a name and the slug matches it, then don't try and
        # reset (i.e.: we only want to set a slug if the name has changed).
        if self.slug and self.name:
            if self.slug == candidate:
                # XXX as long as the instance is pending, as otherwise
                # we skip checking uniqueness on unsaved instances.
                try:
                    insp = inspect(self)
                except sa_exc.NoInspectionAvailable:
                    pass
                else:
                    if insp.persistent or insp.detached:
                        return

        # Iterate until the slug is unique.
        with session.no_autoflush:
            slug = unique(self, self.query, self.__class__.slug, unique_candidate)

        # Finally set the unique slug value.
        self.slug = slug



# -*- coding: utf-8 -*-

"""
Base model container.

Provides a base model container, used by the Pyramid traversal
machinery and a mixin to aid with traversal from an instance
back up the tree.
"""

__all__ = [
    "BaseModelContainer",
    "InstanceTraversalMixin",
]

import re
import logging

from zope.interface import implementer
from zope.interface import alsoProvides

from sqlalchemy.exc import InvalidRequestError

from pyramid.interfaces import ILocation
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import Allow, Deny
from pyramid.security import Authenticated, Everyone

from pyramid_basemodel import Session
from pyramid_basemodel.interfaces import IModelContainer
from pyramid_basemodel.root import BaseRoot

valid_slug = re.compile(r"^[.\w-]{1,64}$", re.U)
logger = logging.getLogger(__name__)


def slug_validator(node, value, regexp=valid_slug):
    """
    Validate slug.

    Defaults to using a slug regexp.
    """
    # Raise a ValueError.
    if not regexp.match(value):
        raise ValueError(f"{value} is not a valid slug.")


@implementer(IModelContainer)
class BaseModelContainer(BaseRoot):
    """Traversal factory that looks up model classes by property."""

    property_name = "slug"
    validation_exception = Exception

    @property
    def _validator(self):
        return slug_validator

    # Default container acl to be private whilst granting authenticated
    # users create permission.
    __acl__ = [
        (Allow, "r:admin", ALL_PERMISSIONS),
        (Allow, Authenticated, "view"),
        (Allow, Authenticated, "create"),
        (Deny, Everyone, ALL_PERMISSIONS),
    ]

    @property
    def name(self):
        """Return plurar version of a class name."""
        return self.model_cls.plural_class_name

    @property
    def class_name(self):
        """Determine class name based on the _class_name or the __tablename__."""
        return self.model_cls.class_name

    @property
    def plural_class_name(self):
        """Return plurar version of a class name."""
        return self.model_cls.plural_class_name

    @property
    def class_slug(self):
        """Class slug based on either _class_slug or __tablename__."""
        return self.model_cls.class_slug

    def get_child(self, key):
        """Query for and return the child instance, if found."""
        column = getattr(self.model_cls, self.property_name)
        query = self.model_cls.query.filter(column == key)
        return query.first()

    def __getitem__(self, key):
        """Lookup model instance by key."""
        try:
            self.validator(None, key)
        except self.validation_exception:
            raise KeyError(key)

        context = self.get_child(key)
        if not context:
            raise KeyError(key)

        return self.locatable(context, key)

    def __init__(self, request, model_cls, key=None, parent=None, **kwargs):
        """Instantiate the container."""
        # Compose.
        if key is None:
            key = model_cls.class_slug
        if parent is None:
            parent = BaseRoot(request)

        self.request = request
        self.model_cls = model_cls
        self.__name__ = key
        self.__parent__ = parent
        if "property_name" in kwargs:
            self.property_name = kwargs.get("property_name")
        if "validator" in kwargs:
            self.validator = kwargs.get("validator")
        else:
            self.validator = self._validator


class InstanceTraversalMixin:
    """Provide a default __parent__ implementation for traversal."""

    request = None
    traversal_key_name = "slug"
    validation_exception = Exception

    @property
    def _validator(self):
        return slug_validator

    @property
    def _base_child_query(self):
        return self.query

    def get_container(self):
        """Reverse up the parent traversal hierarchy until reaching a container."""
        target = self
        while True:
            parent = target.__parent__
            if not parent:
                return None
            if IModelContainer.providedBy(parent):
                return parent
            target = parent

    def locatable(self, context, key, provides=alsoProvides):
        """Make a context object locatable and pass on the request."""
        if not hasattr(context, "__name__"):
            context.__name__ = key
        context._located_parent = self
        context.request = self.request
        if not ILocation.providedBy(context):
            provides(context, ILocation)
        return context

    @property
    def __parent__(self, container_cls=BaseModelContainer, session=Session):
        """Either return ``self.parent``, or a model container object."""
        # If the context has been located, return the container.
        if hasattr(self, "_located_parent"):
            return self._located_parent

        # Add self to the session to avoid ``DetachedInstanceError``s.
        session.add(self)

        # If the model has a parent, return it.
        parent = getattr(self, "parent", None)
        if parent:
            return parent

        # Otherwise instantiate a "fake" traversal container and return that.
        # It's "fake" because it doesn't know about it's parent and doesn't
        # have a copy of the request.
        container = container_cls(None, self.__class__)
        return container

    def __getitem__(self, key):
        """Lookup model instance by key."""
        try:
            self._validator(None, key)
        except self.validation_exception:
            raise KeyError(key)

        # Only lookup children from instances that have them.
        has_children = hasattr(self, "children")
        if not has_children:
            raise KeyError(key)

        # Only lookup if the target column exists.
        column = getattr(self.__class__, self.traversal_key_name, None)
        if not column:
            raise KeyError(key)

        try:
            query = self._base_child_query
            query = query.filter_by(parent=self).filter(column == key)
            context = query.first()
            if not context:
                raise KeyError(key)
        except InvalidRequestError as err:
            # If the query was invalid, the lookup fails, e.g.: if the
            # instance had the requisit properties but they weren't actually
            # sqlalchemy columns.
            logger.warning(err, exc_info=True)
            raise KeyError(key)

        # Return the context, having set the parent and flagged as locatable.
        return self.locatable(context, key)



# -*- coding: utf-8 -*-

"""
Provide a generic model class for storing large binary objects.

To store a bytestring::

  blob = Blob.factory('foo') # sets the blob's name to 'foo'
  blob.value = b'baz'

To store a file::

  f = open('/tmp/foo.pdf')
  blob = Blob.factory('foo', file_like_object=f)
  save(blob)

To store a download from a url::

  blob = Blob.factory('foo')
  blob.update_from_url('http://www.example.com/foo.pdf')
  save(blob)

The Blob's ``value`` is a read only buffer, e.g. to iterate over the contents::

  while True:
      chunk = blob.value.read(1024)
      if chunk is None:
          break
      # do something with ``chunk``

"""

__all__ = [
    "Blob",
]

import logging
from http import HTTPStatus
from io import StringIO

from gzip import GzipFile
from tempfile import NamedTemporaryFile

import requests

from sqlalchemy.schema import Column
from sqlalchemy.types import Unicode
from sqlalchemy.types import LargeBinary

from pyramid_basemodel import Base
from pyramid_basemodel import BaseMixin

logger = logging.getLogger(__name__)


class Blob(Base, BaseMixin):
    """
    Encapsulates a large binary file.

    Instances must have a unique ``self.name``, which has a maximum length
    of 64 characters.

    The binary data is stored in ``self.value``, either directly by assigning
    a bytestring, or passing a file like object to ``self.update()`` or by
    downloading a file from a url using ``self.update_from_url()``. The
    downloaded file can optionally be unzipped if compressed using gzip.

    When reading the data, ``self.value`` is a read only buffer. A convienience
    ``self.get_as_named_tempfile()`` method is provided as an easy way to get
    the value as a readable and writable file that can optionally be closed
    so the data is available from the filesystem.
    """

    __tablename__ = "blobs"

    name = Column(Unicode(64), nullable=False, unique=True)
    value = Column(LargeBinary, nullable=False)

    @classmethod
    def factory(cls, name, file_like_object=None):
        """Create and return."""
        instance = cls()
        instance.update(name, file_like_object=file_like_object)
        return instance

    def update(self, name, file_like_object=None):
        """
        Update value from file like object.

        Update properties, reading the ``file_like_object`` into
        ``self.value`` if provided.
        """
        self.name = name
        if file_like_object is not None:
            self.value = file_like_object.read()

    def update_from_url(self, url):
        """
        Update value from url's content.

        Update ``self.value`` to be the contents of the file downloaded
        from the ``url`` provided.
        """
        # Download the file, raising an exception if the download fails
        # after retrying once.
        attempts = 0
        max_attempts = 2
        while True:
            attempts += 1
            r = requests.get(url)
            if r.status_code == HTTPStatus.OK:
                break
            if attempts < max_attempts:
                continue
            r.raise_for_status()

        self.value = r.content

    def get_as_named_tempfile(self, should_close=False):
        """Read ``self.value`` into and return a named temporary file."""
        # Prepare the temp file.
        f = NamedTemporaryFile(delete=False)

        # Read self.value into it.
        if self.value is not None:
            f.write(self.value)

        # Close the file so its readable from the filename.
        if should_close:
            f.close()

        # Return the file.
        return f

    def __json__(self):
        """Create a JSONable representation."""
        return {"name": self.name}



# -*- coding: utf-8 -*-

"""Shared utility functions for interacting with the data model."""

import os
import logging
from binascii import hexlify

from sqlalchemy import schema

logger = logging.getLogger(__name__)


def generate_random_digest(num_bytes=28, urandom=os.urandom, to_hex=hexlify):
    """
    Generate a random hash and returns the hex digest as a unicode string.

    :param num_bytes: number of bytes to random(select)
    :param urandom: urandom function
    :param to_hex: hexifying function
    """
    # Get random bytes.
    r = urandom(num_bytes)

    # Return as a unicode string.
    return to_hex(r).decode("utf-8")


def ensure_unique(self, query, property_, value, max_iter=30, gen_digest=generate_random_digest):
    """
    Make sure slug is unique.

    Takes a ``candidate`` value for a unique ``property_`` and iterates,
    appending an incremented integer until unique.
    """
    # Unpack
    candidate = value

    # Iterate until the slug is unique.
    n = 0
    n_str = ""
    while True:
        # Keep trying slug, slug-1, slug-2, etc.
        value = f"{candidate}{n_str}"
        existing = None
        existing_instances = query.filter(property_ == value).all()
        for instance in existing_instances:
            if instance != self:
                existing = instance
                break
        if existing and n < 30:
            n += 1
            # If we've tried 1, 2 ... all the way to ``max_iter``, then
            # fallback on appending a random digest rather than a sequential
            # number.
            suffix = str(n) if n < 20 else gen_digest(num_bytes=8)
            n_str = f"-{suffix}"
            continue
        break

    return value


def get_or_create(cls, **kwargs):
    """Get or create a ``cls`` instance using the ``kwargs`` provided."""
    instance = cls.query.filter_by(**kwargs).first()
    if not instance:
        instance = cls(**kwargs)
    return instance


def get_all_matching(cls, column_name, values):
    """
    Return all instances of ``cls`` where ``column_name`` matches one of ``values``.

    :param cls:
    :param column_name:
    :param values:
    """
    column = getattr(cls, column_name)
    query = cls.query.filter(column.in_(values))
    return query.all()


def get_object_id(instance):
    """Return an identifier that's unique across database tables."""
    return f"{instance.__tablename__}#{instance.id}"


def table_args_indexes(tablename, columns):
    """
    Build table indexes.

    Call with a class name and a list of relation id columns to return the
    appropriate op.execute created indexes.

    This is useful as a way to tell `alembic revision --autogenerate` that
    these indexes should exist, even when created manually using `op.execute`.

    Ref: https://bitbucket.org/zzzeek/alembic/issues/233/add-indexes-to-include_object-hook
    """
    indexes = []
    for item in columns:
        if len(item) == 2:
            db_name = item[0]  # db column
            attr_name = item[1]  # sqlalchemy attr
        else:
            db_name = item
            attr_name = item
        idx_name = f"{tablename}_{db_name}_idx"
        idx = schema.Index(idx_name, attr_name)
        indexes.append(idx)
    return tuple(indexes)



# -*- coding: utf-8 -*-

"""
Base traversal root and a mixin class for objects in the Pyramid traversal hierarchy.

Provides a base traversal root and a mixin class for objects in the Pyramid traversal hierarchy.

http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/traversal.html
"""

__all__ = [
    "BaseRoot",
]

import logging

from zope.interface import implementer
from zope.interface import alsoProvides

from pyramid.interfaces import ILocation

logger = logging.getLogger(__name__)


@implementer(ILocation)
class BaseRoot:
    """Base class for traversal factories."""

    __name__ = ""
    __parent__ = None

    def locatable(self, context, key, provides=alsoProvides):
        """Make a context object locatable and return it."""
        if not hasattr(context, "__name__"):
            context.__name__ = key
        context._located_parent = self
        context.request = self.request
        if not ILocation.providedBy(context):
            provides(context, ILocation)
        return context

    def __init__(self, request, key="", parent=None):
        """Initialize BaseRoot class."""
        self.__name__ = key
        self.__parent__ = parent
        self.request = request



# -*- coding: utf-8 -*-

"""Marker interfaces for models and containers."""

__all__ = [
    "IDeclarativeBase",
    "IModel",
    "IModelContainer",
]

from zope.interface import Interface


class IDeclarativeBase(Interface):
    """Implemented by the declarative base and all classes that inherit from it."""


class IModel(IDeclarativeBase):
    """Provided by models."""


class IModelContainer(Interface):
    """Provided by model containers."""



# -*- coding: utf-8 -*-

"""
`BaseContentRoot`` traversal root.

Provide a ``BaseContentRoot`` traversal root for looking up instances
via their containers.
"""

__all__ = [
    "BaseContentRoot",
]

import logging

from zope.interface import Interface
from zope.interface import alsoProvides

from pyramid_basemodel.container import BaseModelContainer
from pyramid_basemodel.root import BaseRoot

logger = logging.getLogger(__name__)


class BaseContentRoot(BaseRoot):
    """Base logic for looking up models."""

    apex = None  # e.g.: (Design, IDesignsContainer, {})
    mapping = {}  # {u'formats': (FileFormat, IFileFormatsContainer, {}), ...}

    def container_factory(
        self, item, key, provides=alsoProvides, default_cls=BaseModelContainer, interface_cls=Interface
    ):
        """Return an instantiated and interface providing container."""
        # Unpack the mapping item.
        model_cls, container_cls_or_interface, kwargs = item

        # If the container_cls_or_interface is an interface, then use the
        # default container cls and mark the instance as providing it.
        is_interface = issubclass(container_cls_or_interface, interface_cls)
        if is_interface:
            container_cls = default_cls
            container_interface = container_cls_or_interface
        else:
            container_cls = container_cls_or_interface

        # Instantiate the model container.
        container = container_cls(self.request, model_cls, key=key, parent=self, **kwargs)

        # Patch it to provide the specific container interface.
        if is_interface:
            provides(container, container_cls_or_interface)

        # Return the container.
        return container

    def __getitem__(self, key):
        """
        Get model from mapping.

        First see if the key is in ``self.mapping``. If it is, return
        a content container configured to look up that model class.
        """
        # If the key matches a traversal container in the mapping, use that.
        if key in self.mapping:
            mapping_item = self.mapping.get(key)
            return self.container_factory(mapping_item, key)

        # Otherwise try and lookup using the apex model class.
        if self.apex:
            container = self.container_factory(self.apex, "")
            return self.locatable(container[key], key)

        raise KeyError(key)
